from dataclasses import dataclass
from enum import Enum
from itertools import count


class WriteMode(Enum):
    ON_EACH = 1
    ON_EACH_NO_START = 2
    ON_EACH_NO_END = 3
    ON_END = 4


class TimeRenderMode(Enum):
    AS_SECONDS = 1
    AS_MILLIS = 2
    AS_UTC = 3
    AS_TIMEZONE = 4


class LogLevel(Enum):
    DEBUG = 1
    INFO = 2
    WARNING = 3
    ERROR = 4
    CRITICAL = 5


_log_level_map = {
    LogLevel.DEBUG: "debug",
    LogLevel.INFO: "info",
    LogLevel.WARNING: "warning",
    LogLevel.ERROR: "error",
    LogLevel.CRITICAL: "critical"
}

_log_num = count(1)


@dataclass
class TallyLogSettings:
    """Class for keeping track of the configuration for a TallyLog"""
    name: str = "custom_{}".format(next(_log_num))  # TODO: USE THIS NAME
    write_mode: WriteMode = WriteMode.ON_EACH
    log_on_init: bool = True
    default_log_level: LogLevel = LogLevel.INFO
    default_expire_time: int = 3600  # in seconds
    time_render: TimeRenderMode = TimeRenderMode.AS_SECONDS
    # TODO: ADD TIMEZONE SUPPORT



class TallyException(Exception):
    pass


class MissingTallyException(TallyException):
    pass


_MISSING_TALLY_ERROR_EXCEPTION_MSG = (
    "No TallySession exists matching your query. "
    "Check your default expire time and try again."
)


class MissingTallyErrorException(MissingTallyException):
    def __init__(self):
        super().__init__(_MISSING_TALLY_ERROR_EXCEPTION_MSG)


_MISSING_TALLY_WARN_EXCEPTION_MSG = (
    "No TallySession exists matching your query. "
    "Check your default expire time, pass `force_create=True`, and try again."
)


class MissingTallyWarnException(MissingTallyException):
    def __init__(self):
        super().__init__(_MISSING_TALLY_WARN_EXCEPTION_MSG)


_DUPLICATE_TALLY_EXCEPTION_MSG = (
    "There are multiple TallySessions matching your query. "
    "Please use a tally id or pass `force_all=True`"
)


class DuplicateTallyException(TallyException):
    def __init__(self):
        super().__init__(_DUPLICATE_TALLY_EXCEPTION_MSG)


_DUPLICATE_NAME_TALLY_EXCEPTION_MSG = (
    "A TallySession with {} already exists. "
    "Please use a unique tally name or pass `force=True` to create it anyway"
)


class DuplicateNameTallyException(TallyException):
    def __init__(self, name: str = "this name"):
        super().__init__(_DUPLICATE_NAME_TALLY_EXCEPTION_MSG.format(name))


_FATAL_TALLY_EXCEPTION_MSG = (
    "A TallyLog object has entered a bad state. "
    "Please check your function calls and try again."
)


class FatalTallyException(TallyException):
    def __init__(self):
        super().__init__(_FATAL_TALLY_EXCEPTION_MSG)


_BAD_ENUM_TALLY_EXCEPTION_MSG = (
    "A bad enum value of {} has been passed to {}. "
    "Please check your function calls and try again."
)


class BadEnumTallyException(TallyException):
    def __init__(self, val: str, enum_type: str):
        super().__init__(_BAD_ENUM_TALLY_EXCEPTION_MSG.format(val, enum_type))



from datetime import datetime, timedelta
from logging import Logger  # trunk-ignore(flake8/TC003)
from typing import Dict, List

# TODO: VERIFY IMPORTS WORK AS PACKAGE
from tallytime.exceptions import (
    BadEnumTallyException,
    DuplicateNameTallyException,
    DuplicateTallyException,
    FatalTallyException,
    MissingTallyErrorException,
    MissingTallyWarnException,
)
from tallytime.loggers import DefaultConsoleLogger
from tallytime.settings import LogLevel, TallyLogSettings, _log_level_map
from tallytime.tally_session import TallySession, _TallySessionId


def _log_message(logger: Logger, level: LogLevel, message: str, *args) -> None:
    if logger is not None:
        log_method_name = _log_level_map.get(level)
        if log_method_name is None:
            raise BadEnumTallyException(level, LogLevel)
        log_method = getattr(logger, log_method_name)

        log_method(message.format(*args))
    else:
        print("NONE")


class TallyLog():
    _sessions: Dict[_TallySessionId, TallySession] = {}
    _name_set: Dict[str, _TallySessionId] = {}
    _settings: TallyLogSettings
    _logger: Logger

    def __init__(self, settings: TallyLogSettings = TallyLogSettings(), logger: Logger = DefaultConsoleLogger()):
        self._settings = settings
        self._logger = logger
        if self._settings.log_on_init:
            _log_message(self._logger, self._settings.default_log_level,
                         "Initialized TallyLog object named {}", self._settings.name)

    def set_settings(self, settings: TallyLogSettings) -> None:
        self._settings = settings

    def set_logger(self, logger: Logger) -> None:
        self._logger = logger

    def __repr__(self):
        return ["{}".format(self._sessions[id]) for id in self._sessions]

    def get_sessions(self) -> List[TallySession]:
        return self._sessions

    def get_session(self, id: _TallySessionId) -> TallySession:
        session = self._sessions.get(id)
        if session is None:
            raise MissingTallyErrorException()
        return session

    def get_sessions_by_name(self, name: str, force_all: bool = False) -> List[TallySession]:
        name_match = self._name_set.get(name, [])

        if len(name_match) == 0:
            raise MissingTallyErrorException()

        elif len(name_match == 1):
            id = name_match[0]
            return [self._session[id]]

        elif (not force_all) and len(name_match) > 1:
            raise DuplicateTallyException()

        sessions = []
        for match in name_match:
            if match not in self._sessions:
                raise FatalTallyException()
            else:
                sessions.append(self._sessions[match])
        return sessions

    # alias
    def get_session_by_name(self, name: str, force_all: bool = False) -> List[TallySession]:
        return self.get_sessions_by_name(name, force_all)

    def _tidy(self) -> None:
        # TODO: UPDATE LOGIC
        items = self._sessions.items()
        for k in range(len(items))[::-1]:
            items = list(items)
            if items[k][1].expire_time < datetime.utcnow():
                del self._sessions[items[k][0]]
            else:
                break

    def register(self, name: str, force: bool = False, start_message: str = None) -> _TallySessionId:
        self._tidy()
        if name in self._name_set:
            if not force:
                raise (DuplicateNameTallyException("name {}".format(name)))

        id = _TallySessionId(name)
        access_token_expires = timedelta(
            seconds=self._settings.default_expire_time)
        expire_time = datetime.utcnow() + access_token_expires
        new_session = TallySession(id, expire_time, start_message)

        self._sessions[id.uuid] = new_session
        if name not in self._name_set:
            self._name_set[id.name] = []
        self._name_set[id.name].append(id)

        return id

    def update(self, id: _TallySessionId, description: str = "Updated", force_create: bool = False) -> _TallySessionId:
        session = self.get_session(id)
        if session is not None:
            session.update(description)
        else:
            if force_create:
                new_session = self.register(id.name)
                # TODO: DEPENDING ON SETTINGS, LOG WARNING AND CREATION
                return new_session
            else:
                raise MissingTallyWarnException()
        # TODO: DEPENDING ON SETTINGS, LOG

        return session.id

    # TODO: ADD HELP TEXT: force makes if not present, also gets first match if present
    def update_by_name(self, name: str, description: str = "Updated", force_create: bool = False, force_all: bool = False) -> List[_TallySessionId]:
        sessions = self.get_sessions_by_name(name, force_all)
        ret = []

        if len(sessions) == 1:
            ret = sessions[0].id
        elif len(sessions) > 1:
            if force_all:
                for session in sessions:
                    session.update(description)
                    ret.append(session.id)
            else:
                raise FatalTallyException()
        elif len(sessions) == 0:
            if force_create:
                ret = [self.register(name)]
            else:
                MissingTallyWarnException()

        # TODO: DEPENDING ON SETTINGS, LOG THIS
        return ret

    def delete(self, id: _TallySessionId) -> TallySession:
        self._tidy()
        session = self.get_session(id)

        if session is None:
            MissingTallyErrorException()
        del self._sessions[id]

        name_set = self._name_set[id.name]
        if len(name_set) == 1:
            del self._name_set[id.name]
        else:
            for index, name_id in enumerate(name_set):
                if name_id == id:
                    del self._name_set[id.name][index]
                    break

        return session

    # TODO: ADD HELP TEXT
    def delete_by_name(self, name: str, force_all: bool = False) -> List[TallySession]:
        self._tidy()
        sessions = self.get_sessions_by_name(name, force_all)
        ret = []

        if len(sessions) == 1:
            ret = [sessions[0]]
        elif len(sessions) > 1:
            if force_all:
                for session in sessions:
                    ret.append(session)
            else:
                raise DuplicateTallyException()
        elif len(sessions) == 0:
            raise MissingTallyErrorException()

        del self._name_set[name]
        for id in [sessions.id for session in sessions]:
            del self._sessions[id]
        return ret

    def display(self, id: _TallySessionId) -> None:
        self._tidy()
        session = self.get_session(id)
        _log_message(
            self._logger, self._settings.default_log_level, str(session))

    def display_and_delete(self, id: _TallySessionId, description: str = "Deleted") -> TallySession:
        self.update(id, description)
        self.display(id)
        return self.delete(id)

    def display_by_name(self, name: str, force_all: bool = False) -> None:
        sessions = self.get_sessions_by_name(name, force_all)
        for session in sessions:
            self.display(session.id)

    def display_and_delete_by_name(self, name: str, description: str = None, force_all: bool = False) -> List[TallySession]:
        sessions = self.get_sessions_by_name(name, force_all)
        ret = []
        for session in sessions:
            ret.append(self.display_and_delete(session.id, description))
        return ret



from typing import List

from tallytime.tally_log import (
    TallyLog,
    TallyLogSettings,
    TallySession,
    _TallySessionId,
)

__all__ = ["tallytime", "set_settings", "set_logger", "get_sessions", "get_session", "get_sessions_by_name", "get_session_by_name", "register",
           "update", "update_by_name", "delete", "delete_by_name", "display", "display_and_delete", "display_by_name", "display_and_delete_by_name"]

_settings = TallyLogSettings()
_settings.name = "tally_root"
tallytime = TallyLog(settings=_settings)


def set_settings(settings: TallyLogSettings) -> None:
    tallytime.set_settings(settings)


def set_logger(logger) -> None:  # TODO: ADD TYPE ANNOTATION
    tallytime.set_logger(logger)


def get_sessions() -> List[TallySession]:
    return tallytime.get_sessions()


def get_session(id: _TallySessionId) -> TallySession:
    return tallytime.get_session(id)


def get_sessions_by_name(name: str, force_all: bool = False) -> List[TallySession]:
    return tallytime.get_sessions_by_name(name, force_all)

# alias


def get_session_by_name(name: str, force_all: bool = False) -> List[TallySession]:
    return tallytime.get_session_by_name(name, force_all)


def register(name: str, force: bool = False, start_message: str = None) -> _TallySessionId:
    return tallytime.register(name, force, start_message)


def update(id: _TallySessionId, description: str = "Updated", force_create: bool = False) -> _TallySessionId:
    return tallytime.update(id, description, force_create)


def update_by_name(name: str, description: str = "Updated", force_create: bool = False, force_all: bool = False) -> List[_TallySessionId]:
    return tallytime.update_by_name(id, description, force_create, force_all)


def delete(id: _TallySessionId) -> TallySession:
    return tallytime.delete(id)


def delete_by_name(name: str, force_all: bool = False) -> List[TallySession]:
    return tallytime.delete_by_name(name, force_all)


def display(id: _TallySessionId) -> None:
    tallytime.display(id)


def display_and_delete(id: _TallySessionId, description: str = "Deleted") -> TallySession:
    return tallytime.display_and_delete(id, description)


def display_by_name(name: str, force_all: bool = False) -> None:
    tallytime.display_by_name(id, force_all)


def display_and_delete_by_name(name: str, description: str = None, force_all: bool = False) -> List[TallySession]:
    return tallytime.display_and_delete_by_name(id, description, force_all)



from logging import Logger, getLogger
from logging.config import dictConfig
from os import makedirs, path


def DefaultConsoleLogger() -> Logger:
    config_dict = {
        "version": 1,
        "formatters": {
            "tally_console_formatter": {
                "format": "%(asctime)s %(levelname)-8s %(name)-14s %(message)s",
                "datefmt": "%m-%d %H:%M:%S",
            }
        },
        "handlers": {
            "tally_console_handler": {
                "class": "logging.StreamHandler",
                "stream": "ext://sys.stdout",
                "formatter": "tally_console_formatter"
            }
        },
        "loggers": {
            "tally_console": {
                "qualname": "tally_console",
                "handlers": ["tally_console_handler"],
                "level": "DEBUG"
            }
        }
    }
    dictConfig(config_dict)
    return getLogger("tally_console")


def DefaultFileLogger(filename: str = "./logs/tallytime.log") -> Logger:
    if not path.exists(path.dirname(filename)):
        makedirs(path.dirname(filename))

    config_dict = {
        "version": 1,
        "formatters": {
            "tally_file_formatter": {
                "format": "%(asctime)s %(levelname)-8s %(name)-14s %(message)s",
                "datefmt": "%m-%d %H:%M:%S",
            }
        },
        "handlers": {
            "tally_file_handler": {
                "()": "logging.FileHandler",
                "filename": filename,
                "formatter": "tally_file_formatter"
            }
        },
        "loggers": {
            "tally_file": {
                "qualname": "tally_file",
                "handlers": ["tally_file_handler"],
                "level": "DEBUG"
            }
        }
    }
    dictConfig(config_dict)
    return getLogger("tally_file")



from dataclasses import dataclass
from datetime import datetime
from typing import List
from uuid import uuid4

from tallytime._utils import _coalesce


@dataclass
class _TallyStatus():
    """Class for a recorded status snapshot, maintained by a TallySession instance"""
    description: str
    time: datetime  # TODO: INVESTIGATE ATOMIC TIME


class _TallySessionId():
    """Class for uniquely identifying a TallySession"""
    name: str
    uuid: uuid4

    def __init__(self, name: str = "None"):
        self.name = name
        self.uuid = uuid4()

    def __repr__(self):
        return _coalesce(self.name, self.uuid.__repr__())

    def __lt__(self, other):
        return self.uuid < other.uuid

    def __eq__(self, other):
        return self.uuid == other

    def __hash__(self):
        return self.uuid.__hash__()


class TallySession():
    """Class for an individual tracker, maintained by a TallyLog instance"""
    id: _TallySessionId
    expire_time: datetime = None
    statuses: List[_TallyStatus] = []
    # TODO: ADD PLUGGABLE REPR SUPPORT

    def __init__(self, id: _TallySessionId, expire_time: datetime, start_message: str = None):
        self.id = id
        self.expire_time = expire_time
        self.statuses = [_TallyStatus(
            _coalesce(start_message, "Initialized"), datetime.utcnow())]

    def update(self, description: str) -> None:
        self.statuses.append(_TallyStatus(description, datetime.utcnow()))

    def __repr__(self):
        ret = "TallySession '{}':\t(Expires at: {})".format(self.id.name,
                                                            self.expire_time)
        for k in range(len(self.statuses)):
            s = self.statuses[k]
            diff = 0 if k == 0 else s.time - self.statuses[k - 1].time
            ret += "\n\t{} \tdiff: {} {} ".format(
                s.description, diff, s.time)

        return ret
